import uuid

from django.core.validators import MinLengthValidator
from django.db import models
from django.utils import timezone


def generate_key():
    return str(uuid.uuid4())[:20]

class ApiEndpoint(models.Model):
    id = models.AutoField(primary_key=True)
    endpoint = models.CharField(max_length=150, null=False, blank=False)


class ApiKey(models.Model):
    id = models.AutoField(primary_key=True)
    name = models.CharField(max_length=128)
    access_key = models.CharField(max_length=128, default=generate_key)
    secret_access_key = models.CharField(max_length=128, default=generate_key)
    updated = models.DateTimeField(null=False, default=timezone.now)
    created = models.DateTimeField(null=False, default=timezone.now)


class ApiPermission(models.Model):
    id = models.AutoField(primary_key=True)
    api_key = models.ForeignKey('ApiKey', null=False, on_delete=models.CASCADE)
    api_endpoint = models.ForeignKey('ApiEndpoint', null=False, on_delete=models.CASCADE)
    methods = models.CharField(max_length=32, null=True, blank=True)


class AuthToken(models.Model):
    id = models.AutoField(primary_key=True)
    key = models.CharField(max_length=80, null=False, default=generate_key)
    auth_type = models.CharField(max_length=150, null=True, blank=True)
    user = models.ForeignKey('User', on_delete=models.CASCADE)
    updated = models.DateTimeField(null=False, default=timezone.now)
    created = models.DateTimeField(null=False, default=timezone.now)


class CognitoUser(models.Model):
    class ErrorMessages:
        NO_SUB_ID_PROVIDED = 'No Cognito sub id provided'

    id = models.AutoField(primary_key=True)
    cognito_sub_id = models.CharField(max_length=36, validators=[MinLengthValidator(36)], unique=True, blank=False,
                                      null=False)

    # have to include this so django will run min_length validator upon save
    def save(self, *args, **kwargs):
        self.full_clean()
        return super(CognitoUser, self).save(*args, **kwargs)


class User(models.Model):
    id = models.AutoField(primary_key=True)
    username = models.CharField(max_length=80, null=True, blank=True)
    password = models.CharField(max_length=128, null=True, blank=True)
    first_name = models.CharField(max_length=100, null=False)
    last_name = models.CharField(max_length=100, null=False)
    is_active = models.BooleanField(null=False, default=True)
    created = models.DateTimeField(null=False, default=timezone.now)
    created_by = models.ForeignKey('User', null=True, blank=True, related_name='users_created', on_delete=models.CASCADE)
    updated = models.DateTimeField(null=False, default=timezone.now)
    updated_by = models.ForeignKey('User', null=True, blank=True, related_name='users_updated', on_delete=models.CASCADE)
    salt = models.CharField(max_length=50, null=True, blank=True)
    password_reset_token = models.CharField(max_length=50, null=True, blank=True)
    password_reset_token_created = models.DateTimeField(null=True, blank=True)
    password_reset_token_expires = models.DateTimeField(null=True, blank=True)
    password_reset_ip = models.GenericIPAddressField(null=True, blank=True)
    failed_login_attempts = models.IntegerField(null=True, blank=True, default=0)
    last_failed_login = models.DateTimeField(null=True, blank=True)

    @property
    def email(self):
        return self.username



import hashlib
import hmac
import io
import json
import time
import unittest
import uuid
from collections import OrderedDict

import bcrypt
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.utils import timezone
from rest_framework.test import APIClient

from rest_framework_signature.settings import auth_settings


def get_hours_in_milliseconds(hours):
    return hours * 60 * 60 * 1000


def get_nonce(timestamp, url, secret_access_key, body=None):
    m = hmac.new(secret_access_key.encode(), digestmod=hashlib.sha1)
    m.update(str(timestamp).encode('utf-8'))
    m.update(url.encode('utf-8'))
    if body:
        sorted_body = sort_body(body)
        sorted_json = json.dumps(sorted_body, ensure_ascii=False).replace(' ', '')
        m.update(sorted_json.encode('utf-8'))
    return m.hexdigest()


def get_timestamp_milliseconds(dt=None):
    if not dt:
        dt = timezone.now()
    # grab time since epoch milliseconds
    return int(time.mktime(dt.timetuple())*1000 + dt.microsecond/1000)


def sort_body(data):
    """
    Where possible, sorts all dictionary and dictionary-like items by key. Reliably sorts
    root-level dictionaries, nested dictionaries, and stringified dictionaries. OrderedDict,
    list, and set types will not be sorted, but the objects they contain are subject to sorting.

    Some edge cases may produce unexpected sorting, including but not limited to:
        - Strings with JSON, list, or set formatted contents will be converted to dict format, with objects
            cast to their inferred types as per json.loads() best judgement
        - OrderedDicts keys won't be sorted, but OrderedDict[key] values will have a sort attempted
        - If contained in a dict, lists and sets will recursively sort each item. Items are not sorted
            within the list, but are sorted within themselves
                Example: sort_body({"a": [{"c": 1, "b": 2}]}) == {"a": [{"b": 2. "c": 1}]}

    :param data: The payload to be sorted before nonce calculation
    :return: Sorted data
    """
    if isinstance(data, dict):
        sorted_result = OrderedDict()
        # Only sort by keys in dictionaries. Keys should be strings, and must all be the same object type.
        #   sorted() fails silently when comparing different types, such as a string and an integer.
        for key, value in sorted(data.items()):
            # everything will be in unicode unless we've converted it and went to the next level of depth
            if isinstance(value, str):
                try:
                    # If we parse the JSON successfully, give the work to a further nested sort_body function
                    parsed_value = json.loads(value)
                    if isinstance(parsed_value, dict) or isinstance(parsed_value, list) or isinstance(parsed_value, set):
                        sorted_result[key] = sort_body(parsed_value)
                    else:  # we only want to use the parsed value if result is dict
                        sorted_result[key] = value
                except ValueError:
                    sorted_result[key] = value
            elif isinstance(value, dict) or isinstance(value, OrderedDict):
                sorted_result[key] = sort_body(value)
            elif isinstance(value, list) or isinstance(value, set):
                # Sorting arrays breaks compatibility, cannot be implemented in drfsig 1.x
                sorted_result[key] = [sort_body(list_item) for list_item in value]
            elif isinstance(value, io.TextIOWrapper) or isinstance(value, InMemoryUploadedFile):
                # There's no sorting to be done with a file.
                continue
            else:
                sorted_result[key] = value
        return sorted_result
    if isinstance(data, list) or isinstance(data, set):
        # Just in case the list contains dictionaries that need sorted
        return [sort_body(list_item) for list_item in data]
    # Anything that isn't a list or dict gets ignored
    return data


def check_valid_reset_token(reset_token, user):
    # check if the reset_token has expired
    expire_time = get_timestamp_milliseconds(user.password_reset_token_expires)
    current_time = get_timestamp_milliseconds()
    is_expired = expire_time < current_time

    if user.password_reset_token != reset_token or is_expired:
        return False

    return True


def generate_email_address():
    return '{0}@{1}.com'.format(str(uuid.uuid4())[:10], str(uuid.uuid4())[:5])


class RestFrameworkSignatureTestClass(unittest.TestCase):

    cognito_enabled = auth_settings.COGNITO_ENABLED
    user_model = auth_settings.get_user_document()
    application_model = auth_settings.get_application_document()
    if not cognito_enabled:
        auth_token_model = auth_settings.get_auth_token_document()

    def get_headers(self, url, body=None):
        headers = self.get_headers_without_auth(url, body)
        if not self.cognito_enabled:
            headers['HTTP_AUTHORIZATION'] = 'Token {0}'.format(self.token.key)
        else:
            headers['HTTP_AUTHORIZATION'] = self.token
        return headers

    def get_headers_without_auth(self, url, body=None):
        timestamp = str(get_timestamp_milliseconds())
        nonce = get_nonce(str(timestamp), url, self.device_token.secret_access_key, body=body)

        headers = {
            auth_settings.TIMESTAMP_HEADER: str(timestamp),
            auth_settings.NONCE_HEADER: nonce,
            auth_settings.API_KEY_HEADER: self.device_token.access_key
        }
        return headers

    def setUp(self):
        # create client to access api endpoints
        self.api_client = APIClient()

        if not self.cognito_enabled:
            # create a user to use to authenticate against
            username = generate_email_address()
            salt = bcrypt.gensalt()
            m = hashlib.sha1()
            m.update('pass1234'.encode('utf-8'))
            sha1_password = m.hexdigest()
            m = hashlib.sha1()
            m.update(sha1_password.encode('utf-8'))
            m.update(salt)
            password = m.hexdigest()
            test_user = self.user_model(
                username=username,
                password=password,
                salt=salt
            )
            test_user.save()
            # create an authentication token
            token, created = self.auth_token_model.objects.get_or_create(user=test_user)
            self.sha1_password = sha1_password
            self.token = token
        else:
            # this is just made up so you will need to mock your cognito user otherwise all will fail
            cognito_sub_id = str(uuid.uuid4())[:30] + str(uuid.uuid4())[:6]
            test_user = self.user_model(
                cognito_sub_id=cognito_sub_id
            )
            test_user.save()
            self.token = str(uuid.uuid4())[:15]

        # create a signature DeviceToken to hash our requests
        access_key = str(uuid.uuid4())[:10]
        secret_access_key = str(uuid.uuid4())[:20]
        device_token = self.application_model(
            name='test-app',
            access_key=access_key,
            secret_access_key=secret_access_key
        )
        device_token.save()
        self.user = test_user
        self.device_token = device_token

    def tearDown(self):
        pass

    class MockRequestsResponse:
        def __init__(self, status_code, text, mock_json_cb=None):
            self.status_code = status_code
            self.text = text
            self.mock_json_cb = mock_json_cb

        def json(self):
            return self.mock_json_cb



class InvalidAuthSettings(Exception):
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(InvalidAuthSettings, self).__init__(message)


import re

import rest_framework.authentication
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import uri_to_iri
from jose.exceptions import JWSError, JWTError
from mongoengine.errors import DoesNotExist
from rest_framework import exceptions

from rest_framework_signature.settings import auth_settings
from rest_framework_signature.helpers import get_nonce, get_timestamp_milliseconds, get_hours_in_milliseconds
from rest_framework_signature.jwt_validator import get_claims, get_client_id_from_access_token
from rest_framework_signature.errors import ErrorMessages


class TokenAuthentication(rest_framework.authentication.BaseAuthentication):
    """
    Simple token based authentication.
    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string "Token ".  For example:
        Authorization: Token 401f7ac837da42b97f613d789819ff93537bee6a
    A custom token model may be used, but must have the following properties.
    * key -- The string identifying the token
    * user -- The user to which the token belongs
    """
    cognito_enabled = auth_settings.COGNITO_ENABLED
    if not cognito_enabled:
        auth_token_model = auth_settings.get_auth_token_document()
    bypass_auth_urls = auth_settings.BYPASS_URLS
    unsecured_urls = auth_settings.UNSECURED_URLS

    def authenticate(self, request):

        # if it is an unsecured url then bypass auth
        request_path = request.path
        if (request.method, request_path) in self.unsecured_urls:
            return

        # first check if they are sending a super key to bypass all auth
        super_key_header = auth_settings.SUPER_KEY_HEADER
        super_key = request.META.get(super_key_header, None) if super_key_header else None
        if super_key and auth_settings.SUPER_KEY_AUTH and super_key == auth_settings.SUPER_KEY_AUTH:
            return AnonymousUser, None

        # first thing check the signature of the request
        self.check_signature(request)

        # special case for posting new users
        if (request.method, request_path) in self.bypass_auth_urls:
            return

        # if they've disabled user auth simply return an anonymous user
        if auth_settings.DISABLE_USER_AUTH:
            return AnonymousUser, None

        try:
            application_model = auth_settings.get_application_document()
            api_key = application_model.objects.get(pk=request.api_key_id)
        except ObjectDoesNotExist:
            api_key = None

        # if specific API keys do not need user auth the bypass it
        if api_key and auth_settings.BYPASS_USER_AUTH_API_KEY_NAMES and \
                api_key.name in auth_settings.BYPASS_USER_AUTH_API_KEY_NAMES:
            return AnonymousUser, None

        # regular authentication
        auth_header = rest_framework.authentication.get_authorization_header(request)
        if not self.cognito_enabled:
            auth = auth_header.split()

            if not auth or auth[0].lower() != b'token':
                raise exceptions.AuthenticationFailed('No Auth Header Present')

            if len(auth) == 1:
                msg = 'Invalid token header. No credentials provided.'
                raise exceptions.AuthenticationFailed(msg)
            elif len(auth) > 2:
                msg = 'Invalid token header. Token string should not contain spaces.'
                raise exceptions.AuthenticationFailed(msg)

            return self.authenticate_credentials(auth[1])
        else:
            if not auth_header:
                raise exceptions.AuthenticationFailed('No Auth Header Present')
            try:
                cognito_client_id = get_client_id_from_access_token(auth_header)
            except (JWSError, JWTError) as ex:
                raise exceptions.AuthenticationFailed('Error decoding token: {0}'.format(ex.args[0]))
            try:
                claims = get_claims(auth_header, cognito_client_id)
            except (JWSError, JWTError) as ex:
                raise exceptions.AuthenticationFailed('Error decoding token: {0}'.format(ex.args[0]))
            return self.authenticate_cognito(claims)

    @staticmethod
    def authenticate_cognito(claims):
        sub = claims.get('sub', None)
        user_model = auth_settings.get_user_document()
        try:
            user = user_model.objects.get(cognito_sub_id=sub)
        except ObjectDoesNotExist:
            raise exceptions.AuthenticationFailed('No user found with cognito sub: {0}'.format(sub))
        else:
            return user, None

    def authenticate_credentials(self, key):
        try:
            token = self.auth_token_model.objects.get(key=key.decode('UTF-8'))
        except (DoesNotExist, ObjectDoesNotExist):
            raise exceptions.AuthenticationFailed('Invalid token')

        if not token.user.is_active:
            raise exceptions.AuthenticationFailed('User inactive or deleted')

        # check if the token has expired
        expire_time = get_hours_in_milliseconds(auth_settings.AUTH_TOKEN_EXPIRATION)
        token_created = get_timestamp_milliseconds(dt=token.updated)
        current_time = get_timestamp_milliseconds()
        if token_created + expire_time < current_time:
            raise exceptions.AuthenticationFailed('Token expired - reauthenticate')
        return token.user, token

    def authenticate_header(self, request):
        return 'Token'

    def check_signature(self, request):
        application_model = auth_settings.get_application_document()
        # collect required headers
        timestamp = request.META.get(auth_settings.TIMESTAMP_HEADER, None)
        nonce = request.META.get(auth_settings.NONCE_HEADER, None)
        api_key = request.META.get(auth_settings.API_KEY_HEADER, None)

        # grab url request is for
        url = uri_to_iri(request.get_full_path())

        # if no api key was passed in raise error
        if not api_key:
            raise exceptions.PermissionDenied(ErrorMessages.PERMISSION_DENIED + ' ' + ErrorMessages.MISSING_API_KEY)

        # get corresponding DeviceToken for request
        try:
            token = application_model.objects.get(access_key=api_key)
        except (DoesNotExist, ObjectDoesNotExist):
            raise exceptions.PermissionDenied(ErrorMessages.PERMISSION_DENIED + ' ' + ErrorMessages.INVALID_API_KEY)

        # set api key id on request
        # todo: move this out of check signature!~~~~~~~~!@@!
        request.api_key_id = token.id

        # ensure that the api key has access to this url
        api_key_has_access = self.validate_api_key(token, url, request.method)
        if not api_key_has_access:
            raise exceptions.PermissionDenied(ErrorMessages.API_KEY_NOT_AUTHORIZED_FOR_ENDPOINT.format(url))

        # to prevent replay attacks if timestamp isn't within last 3 seconds deny request
        current_timestamp = get_timestamp_milliseconds() - auth_settings.REPLAY_ATTACK_TIME
        if not timestamp or int(timestamp) < current_timestamp:
            raise exceptions.PermissionDenied(ErrorMessages.PERMISSION_DENIED + ' ' + ErrorMessages.REPLAY_ERROR)

        if auth_settings.MULTIPART_POST_URLS and url in auth_settings.MULTIPART_POST_URLS:
            return

        # if no nonce was passed in raise error
        if not nonce:
            raise exceptions.PermissionDenied(ErrorMessages.PERMISSION_DENIED + ' ' + ErrorMessages.MISSING_NONCE)

        # if it is a get we are finished
        if request.method in ('GET', 'DELETE'):
            valid_nonce = get_nonce(timestamp, url, token.secret_access_key)
        # post requires us to alphabetize the request data by key
        elif request.method in ('POST', 'PATCH', 'PUT'):
            valid_nonce = get_nonce(timestamp, url, token.secret_access_key, body=request.data)
        else:
            valid_nonce = None

        if valid_nonce != nonce:
            raise exceptions.PermissionDenied(ErrorMessages.PERMISSION_DENIED + ' ' + ErrorMessages.INVALID_NONCE)

    @staticmethod
    def validate_api_key(api_key, url, method):
        if auth_settings.FULL_ACCESS_API_KEY_NAMES and api_key.name in auth_settings.FULL_ACCESS_API_KEY_NAMES:
            return True

        api_permission = auth_settings.get_api_permission_document()
        # grab all permissions
        api_permission.objects.filter(api_key=api_key)
        for api_permission in api_permission.objects.filter(api_key=api_key):
            # check each endpoint to see if url matches the regex
            reg_ex = re.compile(api_permission.api_endpoint.endpoint)
            if reg_ex.match(url):
                # we matched the regex now just see if we have access to that method
                available_methods = api_permission.methods.split(',')
                if method in available_methods:
                    return True
        return False



import hashlib

from django.core.exceptions import ObjectDoesNotExist
from mongoengine.errors import DoesNotExist

from rest_framework_signature.settings import auth_settings


# this allows us to use this backend for authentication
class MongoBackend(object):
    supports_inactive_user = False
    user_model = auth_settings.get_user_document()

    def authenticate(self, username=None, password=None, **kwargs):
        user = self.get_user(username)
        if not user:
            return None
        m = hashlib.sha1()
        m.update(password.encode('utf-8'))
        m.update(user.salt.encode('utf-8'))
        hashed_password = m.hexdigest()
        if user.password == hashed_password:
            return user
        else:
            return None

    def get_user(self, username):
        try:
            return self.user_model.objects(username=username).get()
        except DoesNotExist:
            return None


class MSSQLBackend(object):
    supports_inactive_user = False
    user_model = auth_settings.get_user_document()

    def authenticate(self, username=None, password=None, **kwargs):
        user = self.get_user(username)
        if not user:
            return None
        m = hashlib.sha1()
        m.update(password.encode('utf-8'))
        if type(user.salt) is bytes:
            m.update(user.salt)
        else:
            m.update(user.salt.encode('utf-8'))
        hashed_password = m.hexdigest()
        if user.password == hashed_password:
            return user
        else:
            return None

    def get_user(self, username):
        try:
            return self.user_model.objects.get(username=username)
        except ObjectDoesNotExist:
            return None
